from io import StringIO

from edkrule.interpreter.expression.interpreter import Interpreter
from edkrule.interpreter.lexer.lexer import Lexer
from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.interpreter.parse.parser import Parser


def evals(stringio):
    l = Lexer()
    l.tokenize(stringio)
    es = Parser(l.token_list)
    es.parse()
    exp = Interpreter().interpret(es.statement)
    return exp


def case0():
    rule = StringIO("toDate(a,c)")
    exp = evals(rule)
    assert exp.name == 'toDate'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).text == 'a'
    assert exp.matrix(0).type == TokenType.Identifier
    assert exp.matrix(1).text == 'c'
    assert exp.matrix(1).type == TokenType.Identifier
    # exp.display2([])


def case1():
    rule = StringIO("toDate(a,b,min(c,d))")
    exp = evals(rule)
    assert exp.name == 'toDate'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).text == 'a'
    assert exp.matrix(0).type == TokenType.Identifier
    assert exp.matrix(1).text == 'b'
    assert exp.matrix(1).type == TokenType.Identifier
    assert exp.matrix(2).name == 'min'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).text == 'c'
    assert exp.matrix(2, 0).type == TokenType.Identifier
    assert exp.matrix(2, 1).text == 'd'
    assert exp.matrix(2, 1).type == TokenType.Identifier
    # exp.display2([])


def case2():
    rule = StringIO("1+2+3")
    exp = evals(rule)
    assert exp.name == 'Anonymous'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'Anonymous'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).text == '1'
    assert exp.matrix(0, 0).type == TokenType.RealNumber
    assert exp.matrix(0, 1).text == '+'
    assert exp.matrix(0, 1).type == TokenType.Plus
    assert exp.matrix(0, 2).text == '2'
    assert exp.matrix(0, 2).type == TokenType.RealNumber
    assert exp.matrix(1).text == '+'
    assert exp.matrix(1).type == TokenType.Plus
    assert exp.matrix(2).text == '3'
    assert exp.matrix(2).type == TokenType.RealNumber
    # exp.display2([])


def case3():
    rule = StringIO("a==1+2")
    exp = evals(rule)
    assert exp.name == 'Anonymous'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).text == 'a'
    assert exp.matrix(0).type == TokenType.Identifier
    assert exp.matrix(1).text == '=='
    assert exp.matrix(1).type == TokenType.Eq
    assert exp.matrix(2).name == 'Anonymous'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).text == '1'
    assert exp.matrix(2, 0).type == TokenType.RealNumber
    assert exp.matrix(2, 1).text == '+'
    assert exp.matrix(2, 1).type == TokenType.Plus
    assert exp.matrix(2, 2).text == '2'
    assert exp.matrix(2, 2).type == TokenType.RealNumber
    # exp.display2([])


def case4():
    rule = StringIO("a==1+(2+3)")
    exp = evals(rule)
    assert exp.name == 'Anonymous'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).text == 'a'
    assert exp.matrix(0).type == TokenType.Identifier
    assert exp.matrix(1).text == '=='
    assert exp.matrix(1).type == TokenType.Eq
    assert exp.matrix(2).name == 'Anonymous'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).text == '1'
    assert exp.matrix(2, 0).type == TokenType.RealNumber
    assert exp.matrix(2, 1).text == '+'
    assert exp.matrix(2, 1).type == TokenType.Plus
    assert exp.matrix(2, 2).name == 'Anonymous'
    assert exp.matrix(2, 2).type == TokenType.Expression
    assert exp.matrix(2, 2, 0).name == 'Anonymous'
    assert exp.matrix(2, 2, 0).type == TokenType.Expression
    assert exp.matrix(2, 2, 0, 0).text == '2'
    assert exp.matrix(2, 2, 0, 0).type == TokenType.RealNumber
    assert exp.matrix(2, 2, 0, 1).text == '+'
    assert exp.matrix(2, 2, 0, 1).type == TokenType.Plus
    assert exp.matrix(2, 2, 0, 2).text == '3'
    assert exp.matrix(2, 2, 0, 2).type == TokenType.RealNumber
    # exp.display2([])


def case5():
    rule = StringIO("a==(1+(2+3))")
    exp = evals(rule)
    assert exp.name == 'Anonymous'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).text == 'a'
    assert exp.matrix(0).type == TokenType.Identifier
    assert exp.matrix(1).text == '=='
    assert exp.matrix(1).type == TokenType.Eq
    assert exp.matrix(2).name == 'Anonymous'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).name == 'Anonymous'
    assert exp.matrix(2, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0).text == '1'
    assert exp.matrix(2, 0, 0).type == TokenType.RealNumber
    assert exp.matrix(2, 0, 1).text == '+'
    assert exp.matrix(2, 0, 1).type == TokenType.Plus
    assert exp.matrix(2, 0, 2).name == 'Anonymous'
    assert exp.matrix(2, 0, 2).type == TokenType.Expression
    assert exp.matrix(2, 0, 2, 0).name == 'Anonymous'
    assert exp.matrix(2, 0, 2, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 2, 0, 0).text == '2'
    assert exp.matrix(2, 0, 2, 0, 0).type == TokenType.RealNumber
    assert exp.matrix(2, 0, 2, 0, 1).text == '+'
    assert exp.matrix(2, 0, 2, 0, 1).type == TokenType.Plus
    assert exp.matrix(2, 0, 2, 0, 2).text == '3'
    assert exp.matrix(2, 0, 2, 0, 2).type == TokenType.RealNumber
    # exp.display2([])


def case6():
    rule = StringIO("a==(1+(2+3)) && (1+3+4)")
    exp = evals(rule)
    assert exp.name == 'Anonymous'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'Anonymous'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).text == 'a'
    assert exp.matrix(0, 0).type == TokenType.Identifier
    assert exp.matrix(0, 1).text == '=='
    assert exp.matrix(0, 1).type == TokenType.Eq
    assert exp.matrix(0, 2).name == 'Anonymous'
    assert exp.matrix(0, 2).type == TokenType.Expression
    assert exp.matrix(0, 2, 0).name == 'Anonymous'
    assert exp.matrix(0, 2, 0).type == TokenType.Expression
    assert exp.matrix(0, 2, 0, 0).text == '1'
    assert exp.matrix(0, 2, 0, 0).type == TokenType.RealNumber
    assert exp.matrix(0, 2, 0, 1).text == '+'
    assert exp.matrix(0, 2, 0, 1).type == TokenType.Plus
    assert exp.matrix(0, 2, 0, 2).name == 'Anonymous'
    assert exp.matrix(0, 2, 0, 2).type == TokenType.Expression
    assert exp.matrix(0, 2, 0, 2, 0).name == 'Anonymous'
    assert exp.matrix(0, 2, 0, 2, 0).type == TokenType.Expression
    assert exp.matrix(0, 2, 0, 2, 0, 0).text == '2'
    assert exp.matrix(0, 2, 0, 2, 0, 0).type == TokenType.RealNumber
    assert exp.matrix(0, 2, 0, 2, 0, 1).text == '+'
    assert exp.matrix(0, 2, 0, 2, 0, 1).type == TokenType.Plus
    assert exp.matrix(0, 2, 0, 2, 0, 2).text == '3'
    assert exp.matrix(0, 2, 0, 2, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(1).text == '&&'
    assert exp.matrix(1).type == TokenType.And
    assert exp.matrix(2).name == 'Anonymous'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).name == 'Anonymous'
    assert exp.matrix(2, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0).name == 'Anonymous'
    assert exp.matrix(2, 0, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0, 0).text == '1'
    assert exp.matrix(2, 0, 0, 0).type == TokenType.RealNumber
    assert exp.matrix(2, 0, 0, 1).text == '+'
    assert exp.matrix(2, 0, 0, 1).type == TokenType.Plus
    assert exp.matrix(2, 0, 0, 2).text == '3'
    assert exp.matrix(2, 0, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(2, 0, 1).text == '+'
    assert exp.matrix(2, 0, 1).type == TokenType.Plus
    assert exp.matrix(2, 0, 2).text == '4'
    assert exp.matrix(2, 0, 2).type == TokenType.RealNumber
    # exp.display2([])


def case7():
    rule = StringIO("toDate(a,b,min(c,d)+1+(data(d,e)*5))")
    exp = evals(rule)
    assert exp.name == 'toDate'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).text == 'a'
    assert exp.matrix(0).type == TokenType.Identifier
    assert exp.matrix(1).text == 'b'
    assert exp.matrix(1).type == TokenType.Identifier
    assert exp.matrix(2).name == 'Anonymous'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).name == 'Anonymous'
    assert exp.matrix(2, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0).name == 'min'
    assert exp.matrix(2, 0, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0, 0).text == 'c'
    assert exp.matrix(2, 0, 0, 0).type == TokenType.Identifier
    assert exp.matrix(2, 0, 0, 1).text == 'd'
    assert exp.matrix(2, 0, 0, 1).type == TokenType.Identifier
    assert exp.matrix(2, 0, 1).text == '+'
    assert exp.matrix(2, 0, 1).type == TokenType.Plus
    assert exp.matrix(2, 0, 2).text == '1'
    assert exp.matrix(2, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(2, 1).text == '+'
    assert exp.matrix(2, 1).type == TokenType.Plus
    assert exp.matrix(2, 2).name == 'Anonymous'
    assert exp.matrix(2, 2).type == TokenType.Expression
    assert exp.matrix(2, 2, 0).name == 'Anonymous'
    assert exp.matrix(2, 2, 0).type == TokenType.Expression
    assert exp.matrix(2, 2, 0, 0).name == 'data'
    assert exp.matrix(2, 2, 0, 0).type == TokenType.Expression
    assert exp.matrix(2, 2, 0, 0, 0).text == 'd'
    assert exp.matrix(2, 2, 0, 0, 0).type == TokenType.Identifier
    assert exp.matrix(2, 2, 0, 0, 1).text == 'e'
    assert exp.matrix(2, 2, 0, 0, 1).type == TokenType.Identifier
    assert exp.matrix(2, 2, 0, 1).text == '*'
    assert exp.matrix(2, 2, 0, 1).type == TokenType.Multipy
    assert exp.matrix(2, 2, 0, 2).text == '5'
    assert exp.matrix(2, 2, 0, 2).type == TokenType.RealNumber


def case8():
    rule = StringIO("toDate(a,b,min(c,d))||data(d,e)")
    exp = evals(rule)
    assert exp.name == 'Anonymous'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'toDate'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).text == 'a'
    assert exp.matrix(0, 0).type == TokenType.Identifier
    assert exp.matrix(0, 1).text == 'b'
    assert exp.matrix(0, 1).type == TokenType.Identifier
    assert exp.matrix(0, 2).name == 'min'
    assert exp.matrix(0, 2).type == TokenType.Expression
    assert exp.matrix(0, 2, 0).text == 'c'
    assert exp.matrix(0, 2, 0).type == TokenType.Identifier
    assert exp.matrix(0, 2, 1).text == 'd'
    assert exp.matrix(0, 2, 1).type == TokenType.Identifier
    assert exp.matrix(1).text == '||'
    assert exp.matrix(1).type == TokenType.Or
    assert exp.matrix(2).name == 'data'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).text == 'd'
    assert exp.matrix(2, 0).type == TokenType.Identifier
    assert exp.matrix(2, 1).text == 'e'
    assert exp.matrix(2, 1).type == TokenType.Identifier


def case9():
    rule = StringIO("(1+2)")
    exp = evals(rule)
    assert exp.name == 'Anonymous'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'Anonymous'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).text == '1'
    assert exp.matrix(0, 0).type == TokenType.RealNumber
    assert exp.matrix(0, 1).text == '+'
    assert exp.matrix(0, 1).type == TokenType.Plus
    assert exp.matrix(0, 2).text == '2'
    assert exp.matrix(0, 2).type == TokenType.RealNumber


def case10():
    rule = StringIO("(1+2+(3+4))")
    exp = evals(rule)
    assert exp.name == 'Anonymous'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'Anonymous'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).name == 'Anonymous'
    assert exp.matrix(0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0).text == '1'
    assert exp.matrix(0, 0, 0).type == TokenType.RealNumber
    assert exp.matrix(0, 0, 1).text == '+'
    assert exp.matrix(0, 0, 1).type == TokenType.Plus
    assert exp.matrix(0, 0, 2).text == '2'
    assert exp.matrix(0, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(0, 1).text == '+'
    assert exp.matrix(0, 1).type == TokenType.Plus
    assert exp.matrix(0, 2).name == 'Anonymous'
    assert exp.matrix(0, 2).type == TokenType.Expression
    assert exp.matrix(0, 2, 0).name == 'Anonymous'
    assert exp.matrix(0, 2, 0).type == TokenType.Expression
    assert exp.matrix(0, 2, 0, 0).text == '3'
    assert exp.matrix(0, 2, 0, 0).type == TokenType.RealNumber
    assert exp.matrix(0, 2, 0, 1).text == '+'
    assert exp.matrix(0, 2, 0, 1).type == TokenType.Plus
    assert exp.matrix(0, 2, 0, 2).text == '4'
    assert exp.matrix(0, 2, 0, 2).type == TokenType.RealNumber


def case11():
    rule = StringIO("toDate(a,b,min(c,d))")
    exp = evals(rule)
    assert exp.name == 'toDate'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).text == 'a'
    assert exp.matrix(0).type == TokenType.Identifier
    assert exp.matrix(1).text == 'b'
    assert exp.matrix(1).type == TokenType.Identifier
    assert exp.matrix(2).name == 'min'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).text == 'c'
    assert exp.matrix(2, 0).type == TokenType.Identifier
    assert exp.matrix(2, 1).text == 'd'
    assert exp.matrix(2, 1).type == TokenType.Identifier


def case12():
    rule = StringIO("a==1?true:false")
    exp = evals(rule)
    assert exp.name == 'TernaryOperator'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'Anonymous'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).text == 'a'
    assert exp.matrix(0, 0).type == TokenType.Identifier
    assert exp.matrix(0, 1).text == '=='
    assert exp.matrix(0, 1).type == TokenType.Eq
    assert exp.matrix(0, 2).text == '1'
    assert exp.matrix(0, 2).type == TokenType.RealNumber
    assert exp.matrix(1).text == '?'
    assert exp.matrix(1).type == TokenType.Question
    assert exp.matrix(2).text == 'true'
    assert exp.matrix(2).type == TokenType.TRUE
    assert exp.matrix(3).text == ':'
    assert exp.matrix(3).type == TokenType.Colon
    assert exp.matrix(4).text == 'false'
    assert exp.matrix(4).type == TokenType.FALSE


def case13():
    rule = StringIO("max=a>b?a:b")
    exp = evals(rule)
    assert exp.name == 'Anonymous'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).text == 'max'
    assert exp.matrix(0).type == TokenType.Identifier
    assert exp.matrix(1).text == '='
    assert exp.matrix(1).type == TokenType.Assignment
    assert exp.matrix(2).name == 'TernaryOperator'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).name == 'Anonymous'
    assert exp.matrix(2, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0).text == 'a'
    assert exp.matrix(2, 0, 0).type == TokenType.Identifier
    assert exp.matrix(2, 0, 1).text == '>'
    assert exp.matrix(2, 0, 1).type == TokenType.Gt
    assert exp.matrix(2, 0, 2).text == 'b'
    assert exp.matrix(2, 0, 2).type == TokenType.Identifier
    assert exp.matrix(2, 1).text == '?'
    assert exp.matrix(2, 1).type == TokenType.Question
    assert exp.matrix(2, 2).text == 'a'
    assert exp.matrix(2, 2).type == TokenType.Identifier
    assert exp.matrix(2, 3).text == ':'
    assert exp.matrix(2, 3).type == TokenType.Colon
    assert exp.matrix(2, 4).text == 'b'
    assert exp.matrix(2, 4).type == TokenType.Identifier


def case14():
    rule = StringIO("a==1?true:(x==1?y:z)")
    exp = evals(rule)
    assert exp.name == 'TernaryOperator'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'Anonymous'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).text == 'a'
    assert exp.matrix(0, 0).type == TokenType.Identifier
    assert exp.matrix(0, 1).text == '=='
    assert exp.matrix(0, 1).type == TokenType.Eq
    assert exp.matrix(0, 2).text == '1'
    assert exp.matrix(0, 2).type == TokenType.RealNumber
    assert exp.matrix(1).text == '?'
    assert exp.matrix(1).type == TokenType.Question
    assert exp.matrix(2).text == 'true'
    assert exp.matrix(2).type == TokenType.TRUE
    assert exp.matrix(3).text == ':'
    assert exp.matrix(3).type == TokenType.Colon
    assert exp.matrix(4).name == 'Anonymous'
    assert exp.matrix(4).type == TokenType.Expression
    assert exp.matrix(4, 0).name == 'TernaryOperator'
    assert exp.matrix(4, 0).type == TokenType.Expression
    assert exp.matrix(4, 0, 0).name == 'Anonymous'
    assert exp.matrix(4, 0, 0).type == TokenType.Expression
    assert exp.matrix(4, 0, 0, 0).text == 'x'
    assert exp.matrix(4, 0, 0, 0).type == TokenType.Identifier
    assert exp.matrix(4, 0, 0, 1).text == '=='
    assert exp.matrix(4, 0, 0, 1).type == TokenType.Eq
    assert exp.matrix(4, 0, 0, 2).text == '1'
    assert exp.matrix(4, 0, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(4, 0, 1).text == '?'
    assert exp.matrix(4, 0, 1).type == TokenType.Question
    assert exp.matrix(4, 0, 2).text == 'y'
    assert exp.matrix(4, 0, 2).type == TokenType.Identifier
    assert exp.matrix(4, 0, 3).text == ':'
    assert exp.matrix(4, 0, 3).type == TokenType.Colon
    assert exp.matrix(4, 0, 4).text == 'z'
    assert exp.matrix(4, 0, 4).type == TokenType.Identifier
    # exp.display2([])


def case15():
    rule = StringIO("a==1?(x==1?y:z):false")
    exp = evals(rule)
    assert exp.name == 'TernaryOperator'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'Anonymous'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).text == 'a'
    assert exp.matrix(0, 0).type == TokenType.Identifier
    assert exp.matrix(0, 1).text == '=='
    assert exp.matrix(0, 1).type == TokenType.Eq
    assert exp.matrix(0, 2).text == '1'
    assert exp.matrix(0, 2).type == TokenType.RealNumber
    assert exp.matrix(1).text == '?'
    assert exp.matrix(1).type == TokenType.Question
    assert exp.matrix(2).name == 'Anonymous'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).name == 'TernaryOperator'
    assert exp.matrix(2, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0).name == 'Anonymous'
    assert exp.matrix(2, 0, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0, 0).text == 'x'
    assert exp.matrix(2, 0, 0, 0).type == TokenType.Identifier
    assert exp.matrix(2, 0, 0, 1).text == '=='
    assert exp.matrix(2, 0, 0, 1).type == TokenType.Eq
    assert exp.matrix(2, 0, 0, 2).text == '1'
    assert exp.matrix(2, 0, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(2, 0, 1).text == '?'
    assert exp.matrix(2, 0, 1).type == TokenType.Question
    assert exp.matrix(2, 0, 2).text == 'y'
    assert exp.matrix(2, 0, 2).type == TokenType.Identifier
    assert exp.matrix(2, 0, 3).text == ':'
    assert exp.matrix(2, 0, 3).type == TokenType.Colon
    assert exp.matrix(2, 0, 4).text == 'z'
    assert exp.matrix(2, 0, 4).type == TokenType.Identifier
    assert exp.matrix(3).text == ':'
    assert exp.matrix(3).type == TokenType.Colon
    assert exp.matrix(4).text == 'false'
    assert exp.matrix(4).type == TokenType.FALSE

    # exp.display2([])


def case16():
    rule = StringIO("taDate(a==1?(x==1?y:z):false,a)")
    exp = evals(rule)
    assert exp.name == 'taDate'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'TernaryOperator'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).name == 'Anonymous'
    assert exp.matrix(0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0).text == 'a'
    assert exp.matrix(0, 0, 0).type == TokenType.Identifier
    assert exp.matrix(0, 0, 1).text == '=='
    assert exp.matrix(0, 0, 1).type == TokenType.Eq
    assert exp.matrix(0, 0, 2).text == '1'
    assert exp.matrix(0, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(0, 1).text == '?'
    assert exp.matrix(0, 1).type == TokenType.Question
    assert exp.matrix(0, 2).name == 'Anonymous'
    assert exp.matrix(0, 2).type == TokenType.Expression
    assert exp.matrix(0, 2, 0).name == 'TernaryOperator'
    assert exp.matrix(0, 2, 0).type == TokenType.Expression
    assert exp.matrix(0, 2, 0, 0).name == 'Anonymous'
    assert exp.matrix(0, 2, 0, 0).type == TokenType.Expression
    assert exp.matrix(0, 2, 0, 0, 0).text == 'x'
    assert exp.matrix(0, 2, 0, 0, 0).type == TokenType.Identifier
    assert exp.matrix(0, 2, 0, 0, 1).text == '=='
    assert exp.matrix(0, 2, 0, 0, 1).type == TokenType.Eq
    assert exp.matrix(0, 2, 0, 0, 2).text == '1'
    assert exp.matrix(0, 2, 0, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(0, 2, 0, 1).text == '?'
    assert exp.matrix(0, 2, 0, 1).type == TokenType.Question
    assert exp.matrix(0, 2, 0, 2).text == 'y'
    assert exp.matrix(0, 2, 0, 2).type == TokenType.Identifier
    assert exp.matrix(0, 2, 0, 3).text == ':'
    assert exp.matrix(0, 2, 0, 3).type == TokenType.Colon
    assert exp.matrix(0, 2, 0, 4).text == 'z'
    assert exp.matrix(0, 2, 0, 4).type == TokenType.Identifier
    assert exp.matrix(0, 3).text == ':'
    assert exp.matrix(0, 3).type == TokenType.Colon
    assert exp.matrix(0, 4).text == 'false'
    assert exp.matrix(0, 4).type == TokenType.FALSE
    assert exp.matrix(1).text == 'a'
    assert exp.matrix(1).type == TokenType.Identifier


def case17():
    rule = StringIO("taDate(a==1?(x==1?y:z):false, abc(1+2))==1?a+b+c:true")
    exp = evals(rule)
    assert exp.name == 'TernaryOperator'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'Anonymous'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).name == 'taDate'
    assert exp.matrix(0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0).name == 'TernaryOperator'
    assert exp.matrix(0, 0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0, 0).name == 'Anonymous'
    assert exp.matrix(0, 0, 0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0, 0, 0).text == 'a'
    assert exp.matrix(0, 0, 0, 0, 0).type == TokenType.Identifier
    assert exp.matrix(0, 0, 0, 0, 1).text == '=='
    assert exp.matrix(0, 0, 0, 0, 1).type == TokenType.Eq
    assert exp.matrix(0, 0, 0, 0, 2).text == '1'
    assert exp.matrix(0, 0, 0, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(0, 0, 0, 1).text == '?'
    assert exp.matrix(0, 0, 0, 1).type == TokenType.Question
    assert exp.matrix(0, 0, 0, 2).name == 'Anonymous'
    assert exp.matrix(0, 0, 0, 2).type == TokenType.Expression
    assert exp.matrix(0, 0, 0, 2, 0).name == 'TernaryOperator'
    assert exp.matrix(0, 0, 0, 2, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0, 2, 0, 0).name == 'Anonymous'
    assert exp.matrix(0, 0, 0, 2, 0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0, 2, 0, 0, 0).text == 'x'
    assert exp.matrix(0, 0, 0, 2, 0, 0, 0).type == TokenType.Identifier
    assert exp.matrix(0, 0, 0, 2, 0, 0, 1).text == '=='
    assert exp.matrix(0, 0, 0, 2, 0, 0, 1).type == TokenType.Eq
    assert exp.matrix(0, 0, 0, 2, 0, 0, 2).text == '1'
    assert exp.matrix(0, 0, 0, 2, 0, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(0, 0, 0, 2, 0, 1).text == '?'
    assert exp.matrix(0, 0, 0, 2, 0, 1).type == TokenType.Question
    assert exp.matrix(0, 0, 0, 2, 0, 2).text == 'y'
    assert exp.matrix(0, 0, 0, 2, 0, 2).type == TokenType.Identifier
    assert exp.matrix(0, 0, 0, 2, 0, 3).text == ':'
    assert exp.matrix(0, 0, 0, 2, 0, 3).type == TokenType.Colon
    assert exp.matrix(0, 0, 0, 2, 0, 4).text == 'z'
    assert exp.matrix(0, 0, 0, 2, 0, 4).type == TokenType.Identifier
    assert exp.matrix(0, 0, 0, 3).text == ':'
    assert exp.matrix(0, 0, 0, 3).type == TokenType.Colon
    assert exp.matrix(0, 0, 0, 4).text == 'false'
    assert exp.matrix(0, 0, 0, 4).type == TokenType.FALSE
    assert exp.matrix(0, 0, 1).name == 'abc'
    assert exp.matrix(0, 0, 1).type == TokenType.Expression
    assert exp.matrix(0, 0, 1, 0).name == 'Anonymous'
    assert exp.matrix(0, 0, 1, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 1, 0, 0).text == '1'
    assert exp.matrix(0, 0, 1, 0, 0).type == TokenType.RealNumber
    assert exp.matrix(0, 0, 1, 0, 1).text == '+'
    assert exp.matrix(0, 0, 1, 0, 1).type == TokenType.Plus
    assert exp.matrix(0, 0, 1, 0, 2).text == '2'
    assert exp.matrix(0, 0, 1, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(0, 1).text == '=='
    assert exp.matrix(0, 1).type == TokenType.Eq
    assert exp.matrix(0, 2).text == '1'
    assert exp.matrix(0, 2).type == TokenType.RealNumber
    assert exp.matrix(1).text == '?'
    assert exp.matrix(1).type == TokenType.Question
    assert exp.matrix(2).name == 'Anonymous'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).name == 'Anonymous'
    assert exp.matrix(2, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0).text == 'a'
    assert exp.matrix(2, 0, 0).type == TokenType.Identifier
    assert exp.matrix(2, 0, 1).text == '+'
    assert exp.matrix(2, 0, 1).type == TokenType.Plus
    assert exp.matrix(2, 0, 2).text == 'b'
    assert exp.matrix(2, 0, 2).type == TokenType.Identifier
    assert exp.matrix(2, 1).text == '+'
    assert exp.matrix(2, 1).type == TokenType.Plus
    assert exp.matrix(2, 2).text == 'c'
    assert exp.matrix(2, 2).type == TokenType.Identifier
    assert exp.matrix(3).text == ':'
    assert exp.matrix(3).type == TokenType.Colon
    assert exp.matrix(4).text == 'true'
    assert exp.matrix(4).type == TokenType.TRUE

    # exp.display2([])


def case18():
    rule = StringIO(
        """toDate($C1D1.ONC-392 Administration.ECSTDAT)!=""&& toDate($*.*.LBDAT)!=""&& toDate("2015-1-12 "+$C1D1.ONC-392 Administration.ECSTTIM+":00")!="" &&toDate("2015-1-12 "+$*.*.*+":00")!=""?dateDiff($C1D1.ONC-392 Administration.ECSTDAT+" "+$C1D1.ONC-392 Administration.ECSTTIM+":00",$*.*.LBDAT+" "+$*.*.*+":00","m")>0:true""")
    exp = evals(rule)

    assert exp.name == 'TernaryOperator'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'Anonymous'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).name == 'toDate'
    assert exp.matrix(0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0).text == '$C1D1.ONC-392 Administration.ECSTDAT'
    assert exp.matrix(0, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 1).text == '!='
    assert exp.matrix(0, 1).type == TokenType.NEq
    assert exp.matrix(0, 2).text == '""'
    assert exp.matrix(0, 2).type == TokenType.Identifier
    assert exp.matrix(1).text == '&&'
    assert exp.matrix(1).type == TokenType.And
    assert exp.matrix(2).name == 'Anonymous'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).name == 'toDate'
    assert exp.matrix(2, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0).text == '$*.*.LBDAT'
    assert exp.matrix(2, 0, 0).type == TokenType.Variable
    assert exp.matrix(2, 1).text == '!='
    assert exp.matrix(2, 1).type == TokenType.NEq
    assert exp.matrix(2, 2).text == '""'
    assert exp.matrix(2, 2).type == TokenType.Identifier
    assert exp.matrix(3).text == '&&'
    assert exp.matrix(3).type == TokenType.And
    assert exp.matrix(4).name == 'Anonymous'
    assert exp.matrix(4).type == TokenType.Expression
    assert exp.matrix(4, 0).name == 'toDate'
    assert exp.matrix(4, 0).type == TokenType.Expression
    assert exp.matrix(4, 0, 0).name == 'Anonymous'
    assert exp.matrix(4, 0, 0).type == TokenType.Expression
    assert exp.matrix(4, 0, 0, 0).name == 'Anonymous'
    assert exp.matrix(4, 0, 0, 0).type == TokenType.Expression
    assert exp.matrix(4, 0, 0, 0, 0).text == '"2015-1-12 "'
    assert exp.matrix(4, 0, 0, 0, 0).type == TokenType.Identifier
    assert exp.matrix(4, 0, 0, 0, 1).text == '+'
    assert exp.matrix(4, 0, 0, 0, 1).type == TokenType.Plus
    assert exp.matrix(4, 0, 0, 0, 2).text == '$C1D1.ONC-392 Administration.ECSTTIM'
    assert exp.matrix(4, 0, 0, 0, 2).type == TokenType.Variable
    assert exp.matrix(4, 0, 0, 1).text == '+'
    assert exp.matrix(4, 0, 0, 1).type == TokenType.Plus
    assert exp.matrix(4, 0, 0, 2).text == '":00"'
    assert exp.matrix(4, 0, 0, 2).type == TokenType.Identifier
    assert exp.matrix(4, 1).text == '!='
    assert exp.matrix(4, 1).type == TokenType.NEq
    assert exp.matrix(4, 2).text == '""'
    assert exp.matrix(4, 2).type == TokenType.Identifier
    assert exp.matrix(5).text == '&&'
    assert exp.matrix(5).type == TokenType.And
    assert exp.matrix(6).name == 'Anonymous'
    assert exp.matrix(6).type == TokenType.Expression
    assert exp.matrix(6, 0).name == 'toDate'
    assert exp.matrix(6, 0).type == TokenType.Expression
    assert exp.matrix(6, 0, 0).name == 'Anonymous'
    assert exp.matrix(6, 0, 0).type == TokenType.Expression
    assert exp.matrix(6, 0, 0, 0).name == 'Anonymous'
    assert exp.matrix(6, 0, 0, 0).type == TokenType.Expression
    assert exp.matrix(6, 0, 0, 0, 0).text == '"2015-1-12 "'
    assert exp.matrix(6, 0, 0, 0, 0).type == TokenType.Identifier
    assert exp.matrix(6, 0, 0, 0, 1).text == '+'
    assert exp.matrix(6, 0, 0, 0, 1).type == TokenType.Plus
    assert exp.matrix(6, 0, 0, 0, 2).text == '$*.*.*'
    assert exp.matrix(6, 0, 0, 0, 2).type == TokenType.Variable
    assert exp.matrix(6, 0, 0, 1).text == '+'
    assert exp.matrix(6, 0, 0, 1).type == TokenType.Plus
    assert exp.matrix(6, 0, 0, 2).text == '":00"'
    assert exp.matrix(6, 0, 0, 2).type == TokenType.Identifier
    assert exp.matrix(6, 1).text == '!='
    assert exp.matrix(6, 1).type == TokenType.NEq
    assert exp.matrix(6, 2).text == '""'
    assert exp.matrix(6, 2).type == TokenType.Identifier
    assert exp.matrix(7).text == '?'
    assert exp.matrix(7).type == TokenType.Question
    assert exp.matrix(8).name == 'Anonymous'
    assert exp.matrix(8).type == TokenType.Expression
    assert exp.matrix(8, 0).name == 'dateDiff'
    assert exp.matrix(8, 0).type == TokenType.Expression
    assert exp.matrix(8, 0, 0).name == 'Anonymous'
    assert exp.matrix(8, 0, 0).type == TokenType.Expression
    assert exp.matrix(8, 0, 0, 0).name == 'Anonymous'
    assert exp.matrix(8, 0, 0, 0).type == TokenType.Expression
    assert exp.matrix(8, 0, 0, 0, 0).name == 'Anonymous'
    assert exp.matrix(8, 0, 0, 0, 0).type == TokenType.Expression
    assert exp.matrix(8, 0, 0, 0, 0, 0).text == '$C1D1.ONC-392 Administration.ECSTDAT'
    assert exp.matrix(8, 0, 0, 0, 0, 0).type == TokenType.Variable
    assert exp.matrix(8, 0, 0, 0, 0, 1).text == '+'
    assert exp.matrix(8, 0, 0, 0, 0, 1).type == TokenType.Plus
    assert exp.matrix(8, 0, 0, 0, 0, 2).text == '" "'
    assert exp.matrix(8, 0, 0, 0, 0, 2).type == TokenType.Identifier
    assert exp.matrix(8, 0, 0, 0, 1).text == '+'
    assert exp.matrix(8, 0, 0, 0, 1).type == TokenType.Plus
    assert exp.matrix(8, 0, 0, 0, 2).text == '$C1D1.ONC-392 Administration.ECSTTIM'
    assert exp.matrix(8, 0, 0, 0, 2).type == TokenType.Variable
    assert exp.matrix(8, 0, 0, 1).text == '+'
    assert exp.matrix(8, 0, 0, 1).type == TokenType.Plus
    assert exp.matrix(8, 0, 0, 2).text == '":00"'
    assert exp.matrix(8, 0, 0, 2).type == TokenType.Identifier
    assert exp.matrix(8, 0, 1).name == 'Anonymous'
    assert exp.matrix(8, 0, 1).type == TokenType.Expression
    assert exp.matrix(8, 0, 1, 0).name == 'Anonymous'
    assert exp.matrix(8, 0, 1, 0).type == TokenType.Expression
    assert exp.matrix(8, 0, 1, 0, 0).name == 'Anonymous'
    assert exp.matrix(8, 0, 1, 0, 0).type == TokenType.Expression
    assert exp.matrix(8, 0, 1, 0, 0, 0).text == '$*.*.LBDAT'
    assert exp.matrix(8, 0, 1, 0, 0, 0).type == TokenType.Variable
    assert exp.matrix(8, 0, 1, 0, 0, 1).text == '+'
    assert exp.matrix(8, 0, 1, 0, 0, 1).type == TokenType.Plus
    assert exp.matrix(8, 0, 1, 0, 0, 2).text == '" "'
    assert exp.matrix(8, 0, 1, 0, 0, 2).type == TokenType.Identifier
    assert exp.matrix(8, 0, 1, 0, 1).text == '+'
    assert exp.matrix(8, 0, 1, 0, 1).type == TokenType.Plus
    assert exp.matrix(8, 0, 1, 0, 2).text == '$*.*.*'
    assert exp.matrix(8, 0, 1, 0, 2).type == TokenType.Variable
    assert exp.matrix(8, 0, 1, 1).text == '+'
    assert exp.matrix(8, 0, 1, 1).type == TokenType.Plus
    assert exp.matrix(8, 0, 1, 2).text == '":00"'
    assert exp.matrix(8, 0, 1, 2).type == TokenType.Identifier
    assert exp.matrix(8, 0, 2).text == '"m"'
    assert exp.matrix(8, 0, 2).type == TokenType.Identifier
    assert exp.matrix(8, 1).text == '>'
    assert exp.matrix(8, 1).type == TokenType.Gt
    assert exp.matrix(8, 2).text == '0'
    assert exp.matrix(8, 2).type == TokenType.RealNumber
    assert exp.matrix(9).text == ':'
    assert exp.matrix(9).type == TokenType.Colon
    assert exp.matrix(10).text == 'true'
    assert exp.matrix(10).type == TokenType.TRUE
    # exp.display2([])


def case19():
    rule = StringIO(
        """autoValue(RoundN(sum(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG, $*.Target Lesions Assessment (Details) (Screening).TLLOC != 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU == "CM") , multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG, $*.Target Lesions Assessment (Details) (Screening).TLLOC != 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU == "MM"),1/10) , getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT, $*.Target Lesions Assessment (Details) (Screening).TLLOC == 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU== "CM") , multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT, $*.Target Lesions Assessment (Details) (Screening).TLLOC == 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU== "MM"),1/10)), 0.01), true)""")
    exp = evals(rule)

    assert exp.name == 'autoValue'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'RoundN'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).name == 'sum'
    assert exp.matrix(0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0).name == 'getSumOfItemInLog'
    assert exp.matrix(0, 0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0, 0).text == '$*.Target Lesions Assessment (Details) (Screening).TLLONG'
    assert exp.matrix(0, 0, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 0, 1).name == 'Anonymous'
    assert exp.matrix(0, 0, 0, 1).type == TokenType.Expression
    assert exp.matrix(0, 0, 0, 1, 0).name == 'Anonymous'
    assert exp.matrix(0, 0, 0, 1, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0, 1, 0, 0).text == '$*.Target Lesions Assessment (Details) (Screening).TLLOC '
    assert exp.matrix(0, 0, 0, 1, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 0, 1, 0, 1).text == '!='
    assert exp.matrix(0, 0, 0, 1, 0, 1).type == TokenType.NEq
    assert exp.matrix(0, 0, 0, 1, 0, 2).text == '11'
    assert exp.matrix(0, 0, 0, 1, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(0, 0, 0, 1, 1).text == '&&'
    assert exp.matrix(0, 0, 0, 1, 1).type == TokenType.And
    assert exp.matrix(0, 0, 0, 1, 2).name == 'Anonymous'
    assert exp.matrix(0, 0, 0, 1, 2).type == TokenType.Expression
    assert exp.matrix(0, 0, 0, 1, 2, 0).text == '$*.Target Lesions Assessment (Details) (Screening).TLDIAU '
    assert exp.matrix(0, 0, 0, 1, 2, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 0, 1, 2, 1).text == '=='
    assert exp.matrix(0, 0, 0, 1, 2, 1).type == TokenType.Eq
    assert exp.matrix(0, 0, 0, 1, 2, 2).text == '"CM"'
    assert exp.matrix(0, 0, 0, 1, 2, 2).type == TokenType.Identifier
    assert exp.matrix(0, 0, 1).name == 'multiply'
    assert exp.matrix(0, 0, 1).type == TokenType.Expression
    assert exp.matrix(0, 0, 1, 0).name == 'getSumOfItemInLog'
    assert exp.matrix(0, 0, 1, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 1, 0, 0).text == '$*.Target Lesions Assessment (Details) (Screening).TLLONG'
    assert exp.matrix(0, 0, 1, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 1, 0, 1).name == 'Anonymous'
    assert exp.matrix(0, 0, 1, 0, 1).type == TokenType.Expression
    assert exp.matrix(0, 0, 1, 0, 1, 0).name == 'Anonymous'
    assert exp.matrix(0, 0, 1, 0, 1, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 1, 0, 1, 0, 0).text == '$*.Target Lesions Assessment (Details) (Screening).TLLOC '
    assert exp.matrix(0, 0, 1, 0, 1, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 1, 0, 1, 0, 1).text == '!='
    assert exp.matrix(0, 0, 1, 0, 1, 0, 1).type == TokenType.NEq
    assert exp.matrix(0, 0, 1, 0, 1, 0, 2).text == '11'
    assert exp.matrix(0, 0, 1, 0, 1, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(0, 0, 1, 0, 1, 1).text == '&&'
    assert exp.matrix(0, 0, 1, 0, 1, 1).type == TokenType.And
    assert exp.matrix(0, 0, 1, 0, 1, 2).name == 'Anonymous'
    assert exp.matrix(0, 0, 1, 0, 1, 2).type == TokenType.Expression
    assert exp.matrix(0, 0, 1, 0, 1, 2, 0).text == '$*.Target Lesions Assessment (Details) (Screening).TLDIAU '
    assert exp.matrix(0, 0, 1, 0, 1, 2, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 1, 0, 1, 2, 1).text == '=='
    assert exp.matrix(0, 0, 1, 0, 1, 2, 1).type == TokenType.Eq
    assert exp.matrix(0, 0, 1, 0, 1, 2, 2).text == '"MM"'
    assert exp.matrix(0, 0, 1, 0, 1, 2, 2).type == TokenType.Identifier
    assert exp.matrix(0, 0, 1, 1).name == 'Anonymous'
    assert exp.matrix(0, 0, 1, 1).type == TokenType.Expression
    assert exp.matrix(0, 0, 1, 1, 0).text == '1'
    assert exp.matrix(0, 0, 1, 1, 0).type == TokenType.RealNumber
    assert exp.matrix(0, 0, 1, 1, 1).text == '/'
    assert exp.matrix(0, 0, 1, 1, 1).type == TokenType.Divide
    assert exp.matrix(0, 0, 1, 1, 2).text == '10'
    assert exp.matrix(0, 0, 1, 1, 2).type == TokenType.RealNumber
    assert exp.matrix(0, 0, 2).name == 'getSumOfItemInLog'
    assert exp.matrix(0, 0, 2).type == TokenType.Expression
    assert exp.matrix(0, 0, 2, 0).text == '$*.Target Lesions Assessment (Details) (Screening).TLSHORT'
    assert exp.matrix(0, 0, 2, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 2, 1).name == 'Anonymous'
    assert exp.matrix(0, 0, 2, 1).type == TokenType.Expression
    assert exp.matrix(0, 0, 2, 1, 0).name == 'Anonymous'
    assert exp.matrix(0, 0, 2, 1, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 2, 1, 0, 0).text == '$*.Target Lesions Assessment (Details) (Screening).TLLOC '
    assert exp.matrix(0, 0, 2, 1, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 2, 1, 0, 1).text == '=='
    assert exp.matrix(0, 0, 2, 1, 0, 1).type == TokenType.Eq
    assert exp.matrix(0, 0, 2, 1, 0, 2).text == '11'
    assert exp.matrix(0, 0, 2, 1, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(0, 0, 2, 1, 1).text == '&&'
    assert exp.matrix(0, 0, 2, 1, 1).type == TokenType.And
    assert exp.matrix(0, 0, 2, 1, 2).name == 'Anonymous'
    assert exp.matrix(0, 0, 2, 1, 2).type == TokenType.Expression
    assert exp.matrix(0, 0, 2, 1, 2, 0).text == '$*.Target Lesions Assessment (Details) (Screening).TLDIAU'
    assert exp.matrix(0, 0, 2, 1, 2, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 2, 1, 2, 1).text == '=='
    assert exp.matrix(0, 0, 2, 1, 2, 1).type == TokenType.Eq
    assert exp.matrix(0, 0, 2, 1, 2, 2).text == '"CM"'
    assert exp.matrix(0, 0, 2, 1, 2, 2).type == TokenType.Identifier
    assert exp.matrix(0, 0, 3).name == 'multiply'
    assert exp.matrix(0, 0, 3).type == TokenType.Expression
    assert exp.matrix(0, 0, 3, 0).name == 'getSumOfItemInLog'
    assert exp.matrix(0, 0, 3, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 3, 0, 0).text == '$*.Target Lesions Assessment (Details) (Screening).TLSHORT'
    assert exp.matrix(0, 0, 3, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 3, 0, 1).name == 'Anonymous'
    assert exp.matrix(0, 0, 3, 0, 1).type == TokenType.Expression
    assert exp.matrix(0, 0, 3, 0, 1, 0).name == 'Anonymous'
    assert exp.matrix(0, 0, 3, 0, 1, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 3, 0, 1, 0, 0).text == '$*.Target Lesions Assessment (Details) (Screening).TLLOC '
    assert exp.matrix(0, 0, 3, 0, 1, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 3, 0, 1, 0, 1).text == '=='
    assert exp.matrix(0, 0, 3, 0, 1, 0, 1).type == TokenType.Eq
    assert exp.matrix(0, 0, 3, 0, 1, 0, 2).text == '11'
    assert exp.matrix(0, 0, 3, 0, 1, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(0, 0, 3, 0, 1, 1).text == '&&'
    assert exp.matrix(0, 0, 3, 0, 1, 1).type == TokenType.And
    assert exp.matrix(0, 0, 3, 0, 1, 2).name == 'Anonymous'
    assert exp.matrix(0, 0, 3, 0, 1, 2).type == TokenType.Expression
    assert exp.matrix(0, 0, 3, 0, 1, 2, 0).text == '$*.Target Lesions Assessment (Details) (Screening).TLDIAU'
    assert exp.matrix(0, 0, 3, 0, 1, 2, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 3, 0, 1, 2, 1).text == '=='
    assert exp.matrix(0, 0, 3, 0, 1, 2, 1).type == TokenType.Eq
    assert exp.matrix(0, 0, 3, 0, 1, 2, 2).text == '"MM"'
    assert exp.matrix(0, 0, 3, 0, 1, 2, 2).type == TokenType.Identifier
    assert exp.matrix(0, 0, 3, 1).name == 'Anonymous'
    assert exp.matrix(0, 0, 3, 1).type == TokenType.Expression
    assert exp.matrix(0, 0, 3, 1, 0).text == '1'
    assert exp.matrix(0, 0, 3, 1, 0).type == TokenType.RealNumber
    assert exp.matrix(0, 0, 3, 1, 1).text == '/'
    assert exp.matrix(0, 0, 3, 1, 1).type == TokenType.Divide
    assert exp.matrix(0, 0, 3, 1, 2).text == '10'
    assert exp.matrix(0, 0, 3, 1, 2).type == TokenType.RealNumber
    assert exp.matrix(0, 1).text == '0.01'
    assert exp.matrix(0, 1).type == TokenType.RealNumber
    assert exp.matrix(1).text == 'true'
    assert exp.matrix(1).type == TokenType.TRUE


def case20():
    rule = StringIO(
        """mustAnswer($*.*.*)&&(getICFVersion($*.Date of Visit.SVDAT, $*.Informed Consent.DSSTDAT, $*.Informed Consent.VERSION)=='V1.5'||getICFVersion($*.Date of Visit.SVDAT, $*.Informed Consent.DSSTDAT, $*.Informed Consent.VERSION)=='V1.6')?($*.*.*=='0'||$*.*.*=='1'||$*.*.*=='2'):true""")
    exp = evals(rule)

    assert exp.name == 'TernaryOperator'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'Anonymous'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).name == 'mustAnswer'
    assert exp.matrix(0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0).text == '$*.*.*'
    assert exp.matrix(0, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 1).text == '&&'
    assert exp.matrix(0, 1).type == TokenType.And
    assert exp.matrix(0, 2).name == 'Anonymous'
    assert exp.matrix(0, 2).type == TokenType.Expression
    assert exp.matrix(0, 2, 0).name == 'Anonymous'
    assert exp.matrix(0, 2, 0).type == TokenType.Expression
    assert exp.matrix(0, 2, 0, 0).name == 'Anonymous'
    assert exp.matrix(0, 2, 0, 0).type == TokenType.Expression
    assert exp.matrix(0, 2, 0, 0, 0).name == 'getICFVersion'
    assert exp.matrix(0, 2, 0, 0, 0).type == TokenType.Expression
    assert exp.matrix(0, 2, 0, 0, 0, 0).text == '$*.Date of Visit.SVDAT'
    assert exp.matrix(0, 2, 0, 0, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 2, 0, 0, 0, 1).text == '$*.Informed Consent.DSSTDAT'
    assert exp.matrix(0, 2, 0, 0, 0, 1).type == TokenType.Variable
    assert exp.matrix(0, 2, 0, 0, 0, 2).text == '$*.Informed Consent.VERSION'
    assert exp.matrix(0, 2, 0, 0, 0, 2).type == TokenType.Variable
    assert exp.matrix(0, 2, 0, 0, 1).text == '=='
    assert exp.matrix(0, 2, 0, 0, 1).type == TokenType.Eq
    assert exp.matrix(0, 2, 0, 0, 2).text == "'V1.5'"
    assert exp.matrix(0, 2, 0, 0, 2).type == TokenType.Identifier
    assert exp.matrix(0, 2, 0, 1).text == '||'
    assert exp.matrix(0, 2, 0, 1).type == TokenType.Or
    assert exp.matrix(0, 2, 0, 2).name == 'Anonymous'
    assert exp.matrix(0, 2, 0, 2).type == TokenType.Expression
    assert exp.matrix(0, 2, 0, 2, 0).name == 'getICFVersion'
    assert exp.matrix(0, 2, 0, 2, 0).type == TokenType.Expression
    assert exp.matrix(0, 2, 0, 2, 0, 0).text == '$*.Date of Visit.SVDAT'
    assert exp.matrix(0, 2, 0, 2, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 2, 0, 2, 0, 1).text == '$*.Informed Consent.DSSTDAT'
    assert exp.matrix(0, 2, 0, 2, 0, 1).type == TokenType.Variable
    assert exp.matrix(0, 2, 0, 2, 0, 2).text == '$*.Informed Consent.VERSION'
    assert exp.matrix(0, 2, 0, 2, 0, 2).type == TokenType.Variable
    assert exp.matrix(0, 2, 0, 2, 1).text == '=='
    assert exp.matrix(0, 2, 0, 2, 1).type == TokenType.Eq
    assert exp.matrix(0, 2, 0, 2, 2).text == "'V1.6'"
    assert exp.matrix(0, 2, 0, 2, 2).type == TokenType.Identifier
    assert exp.matrix(1).text == '?'
    assert exp.matrix(1).type == TokenType.Question
    assert exp.matrix(2).name == 'Anonymous'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).name == 'Anonymous'
    assert exp.matrix(2, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0).name == 'Anonymous'
    assert exp.matrix(2, 0, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0, 0).name == 'Anonymous'
    assert exp.matrix(2, 0, 0, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0, 0, 0).text == '$*.*.*'
    assert exp.matrix(2, 0, 0, 0, 0).type == TokenType.Variable
    assert exp.matrix(2, 0, 0, 0, 1).text == '=='
    assert exp.matrix(2, 0, 0, 0, 1).type == TokenType.Eq
    assert exp.matrix(2, 0, 0, 0, 2).text == "'0'"
    assert exp.matrix(2, 0, 0, 0, 2).type == TokenType.Identifier
    assert exp.matrix(2, 0, 0, 1).text == '||'
    assert exp.matrix(2, 0, 0, 1).type == TokenType.Or
    assert exp.matrix(2, 0, 0, 2).name == 'Anonymous'
    assert exp.matrix(2, 0, 0, 2).type == TokenType.Expression
    assert exp.matrix(2, 0, 0, 2, 0).text == '$*.*.*'
    assert exp.matrix(2, 0, 0, 2, 0).type == TokenType.Variable
    assert exp.matrix(2, 0, 0, 2, 1).text == '=='
    assert exp.matrix(2, 0, 0, 2, 1).type == TokenType.Eq
    assert exp.matrix(2, 0, 0, 2, 2).text == "'1'"
    assert exp.matrix(2, 0, 0, 2, 2).type == TokenType.Identifier
    assert exp.matrix(2, 0, 1).text == '||'
    assert exp.matrix(2, 0, 1).type == TokenType.Or
    assert exp.matrix(2, 0, 2).name == 'Anonymous'
    assert exp.matrix(2, 0, 2).type == TokenType.Expression
    assert exp.matrix(2, 0, 2, 0).text == '$*.*.*'
    assert exp.matrix(2, 0, 2, 0).type == TokenType.Variable
    assert exp.matrix(2, 0, 2, 1).text == '=='
    assert exp.matrix(2, 0, 2, 1).type == TokenType.Eq
    assert exp.matrix(2, 0, 2, 2).text == "'2'"
    assert exp.matrix(2, 0, 2, 2).type == TokenType.Identifier
    assert exp.matrix(3).text == ':'
    assert exp.matrix(3).type == TokenType.Colon
    assert exp.matrix(4).text == 'true'
    assert exp.matrix(4).type == TokenType.TRUE


def case21():
    rule = StringIO(
        """toDate($*.*.*)!=""&&toDate(getMinByLog($*.Informed Consent.DSSTDAT))!=""?dateDiff(getMinByLog($*.Informed Consent.DSSTDAT),$*.*.*, "D")>0:true""")
    exp = evals(rule)
    assert exp.name == 'TernaryOperator'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'Anonymous'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).name == 'toDate'
    assert exp.matrix(0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0).text == '$*.*.*'
    assert exp.matrix(0, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 1).text == '!='
    assert exp.matrix(0, 1).type == TokenType.NEq
    assert exp.matrix(0, 2).text == '""'
    assert exp.matrix(0, 2).type == TokenType.Identifier
    assert exp.matrix(1).text == '&&'
    assert exp.matrix(1).type == TokenType.And
    assert exp.matrix(2).name == 'Anonymous'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).name == 'toDate'
    assert exp.matrix(2, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0).name == 'getMinByLog'
    assert exp.matrix(2, 0, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0, 0).text == '$*.Informed Consent.DSSTDAT'
    assert exp.matrix(2, 0, 0, 0).type == TokenType.Variable
    assert exp.matrix(2, 1).text == '!='
    assert exp.matrix(2, 1).type == TokenType.NEq
    assert exp.matrix(2, 2).text == '""'
    assert exp.matrix(2, 2).type == TokenType.Identifier
    assert exp.matrix(3).text == '?'
    assert exp.matrix(3).type == TokenType.Question
    assert exp.matrix(4).name == 'Anonymous'
    assert exp.matrix(4).type == TokenType.Expression
    assert exp.matrix(4, 0).name == 'dateDiff'
    assert exp.matrix(4, 0).type == TokenType.Expression
    assert exp.matrix(4, 0, 0).name == 'getMinByLog'
    assert exp.matrix(4, 0, 0).type == TokenType.Expression
    assert exp.matrix(4, 0, 0, 0).text == '$*.Informed Consent.DSSTDAT'
    assert exp.matrix(4, 0, 0, 0).type == TokenType.Variable
    assert exp.matrix(4, 0, 1).text == '$*.*.*'
    assert exp.matrix(4, 0, 1).type == TokenType.Variable
    assert exp.matrix(4, 0, 2).text == '"D"'
    assert exp.matrix(4, 0, 2).type == TokenType.Identifier
    assert exp.matrix(4, 1).text == '>'
    assert exp.matrix(4, 1).type == TokenType.Gt
    assert exp.matrix(4, 2).text == '0'
    assert exp.matrix(4, 2).type == TokenType.RealNumber
    assert exp.matrix(5).text == ':'
    assert exp.matrix(5).type == TokenType.Colon
    assert exp.matrix(6).text == 'true'
    assert exp.matrix(6).type == TokenType.TRUE


def case22():
    rule = StringIO("""isValidDate($*.*.*)""")
    exp = evals(rule)
    assert exp.name == 'isValidDate'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).text == '$*.*.*'
    assert exp.matrix(0).type == TokenType.Variable


def case23():
    rule = StringIO(
        """autoValue((toNum($*.*.EGORRES.881)!=''&&toNum($*.*.EGORRES.879)!=''?RoundN(toNum($*.*.EGORRES.879)/Math.pow((toNum($*.*.EGORRES.881)/1000), 0.33), "1"):""), true)""")
    exp = evals(rule)
    assert exp.name == 'autoValue'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'Anonymous'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).name == 'TernaryOperator'
    assert exp.matrix(0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0).name == 'Anonymous'
    assert exp.matrix(0, 0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0, 0).name == 'toNum'
    assert exp.matrix(0, 0, 0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0, 0, 0).text == '$*.*.EGORRES.881'
    assert exp.matrix(0, 0, 0, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 0, 1).text == '!='
    assert exp.matrix(0, 0, 0, 1).type == TokenType.NEq
    assert exp.matrix(0, 0, 0, 2).text == '\'\''
    assert exp.matrix(0,0,0,2).type == TokenType.Identifier
    assert exp.matrix(0,0,1).text == '&&'
    assert exp.matrix(0,0,1).type == TokenType.And
    assert exp.matrix(0,0,2).name == 'Anonymous'
    assert exp.matrix(0,0,2).type == TokenType.Expression
    assert exp.matrix(0,0,2,0).name == 'toNum'
    assert exp.matrix(0,0,2,0).type == TokenType.Expression
    assert exp.matrix(0,0,2,0,0).text == '$*.*.EGORRES.879'
    assert exp.matrix(0,0,2,0,0).type == TokenType.Variable
    assert exp.matrix(0,0,2,1).text == '!='
    assert exp.matrix(0,0,2,1).type == TokenType.NEq
    assert exp.matrix(0,0,2,2).text == '\'\''
    assert exp.matrix(0, 0, 2, 2).type == TokenType.Identifier
    assert exp.matrix(0, 0, 3).text == '?'
    assert exp.matrix(0, 0, 3).type == TokenType.Question
    assert exp.matrix(0, 0, 4).name == 'RoundN'
    assert exp.matrix(0, 0, 4).type == TokenType.Expression
    assert exp.matrix(0, 0, 4, 0).name == 'Anonymous'
    assert exp.matrix(0, 0, 4, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 4, 0, 0).name == 'toNum'
    assert exp.matrix(0, 0, 4, 0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 4, 0, 0, 0).text == '$*.*.EGORRES.879'
    assert exp.matrix(0, 0, 4, 0, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 4, 0, 1).text == '/'
    assert exp.matrix(0, 0, 4, 0, 1).type == TokenType.Divide
    assert exp.matrix(0, 0, 4, 0, 2).name == 'Math.pow'
    assert exp.matrix(0, 0, 4, 0, 2).type == TokenType.Expression
    assert exp.matrix(0, 0, 4, 0, 2, 0).name == 'Anonymous'
    assert exp.matrix(0, 0, 4, 0, 2, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 4, 0, 2, 0, 0).name == 'Anonymous'
    assert exp.matrix(0, 0, 4, 0, 2, 0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 4, 0, 2, 0, 0, 0).name == 'toNum'
    assert exp.matrix(0, 0, 4, 0, 2, 0, 0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 4, 0, 2, 0, 0, 0, 0).text == '$*.*.EGORRES.881'
    assert exp.matrix(0, 0, 4, 0, 2, 0, 0, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 4, 0, 2, 0, 0, 1).text == '/'
    assert exp.matrix(0, 0, 4, 0, 2, 0, 0, 1).type == TokenType.Divide
    assert exp.matrix(0, 0, 4, 0, 2, 0, 0, 2).text == '1000'
    assert exp.matrix(0, 0, 4, 0, 2, 0, 0, 2).type == TokenType.RealNumber
    assert exp.matrix(0, 0, 4, 0, 2, 1).text == '0.33'
    assert exp.matrix(0, 0, 4, 0, 2, 1).type == TokenType.RealNumber
    assert exp.matrix(0, 0, 4, 1).text == '"1"'
    assert exp.matrix(0, 0, 4, 1).type == TokenType.Identifier
    assert exp.matrix(0, 0, 5).text == ':'
    assert exp.matrix(0, 0, 5).type == TokenType.Colon
    assert exp.matrix(0, 0, 6).text == '""'
    assert exp.matrix(0, 0, 6).type == TokenType.Identifier
    assert exp.matrix(1).text == 'true'
    assert exp.matrix(1).type == TokenType.TRUE

 


def case24():
    rule = StringIO("$*.*.EGPERF==\"Y\"")
    exp = evals(rule)
    assert exp.name == 'Anonymous'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).text == '$*.*.EGPERF'
    assert exp.matrix(0).type == TokenType.Variable
    assert exp.matrix(1).text == '=='
    assert exp.matrix(1).type == TokenType.Eq
    assert exp.matrix(2).text == '"Y"'
    assert exp.matrix(2).type == TokenType.Identifier


def case25():
    rule = StringIO("$*.*.EGORRES.12394==\"CS\"")
    exp = evals(rule)
    assert exp.name == 'Anonymous'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).text == '$*.*.EGORRES.12394'
    assert exp.matrix(0).type == TokenType.Variable
    assert exp.matrix(1).text == '=='
    assert exp.matrix(1).type == TokenType.Eq
    assert exp.matrix(2).text == '"CS"'
    assert exp.matrix(2).type == TokenType.Identifier


def case26():
    rule = StringIO("""condition(mustAnswer(toNum($*.*.*)), isRange($*.*.*, 40,160))""")
    exp = evals(rule)
    assert exp.name == 'condition'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'mustAnswer'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).name == 'toNum'
    assert exp.matrix(0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0).text == '$*.*.*'
    assert exp.matrix(0, 0, 0).type == TokenType.Variable
    assert exp.matrix(1).name == 'isRange'
    assert exp.matrix(1).type == TokenType.Expression
    assert exp.matrix(1, 0).text == '$*.*.*'
    assert exp.matrix(1, 0).type == TokenType.Variable
    assert exp.matrix(1, 1).text == '40'
    assert exp.matrix(1, 1).type == TokenType.RealNumber
    assert exp.matrix(1, 2).text == '160'
    assert exp.matrix(1, 2).type == TokenType.RealNumber


def case27():
    rule = StringIO("""mustAnswer($*.*.CRONGO)&&$*.*.CRONGO=='Y'?$*.*.*=='':true""")
    exp = evals(rule)
    assert exp.name == 'TernaryOperator'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'mustAnswer'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).text == '$*.*.CRONGO'
    assert exp.matrix(0, 0).type == TokenType.Variable
    assert exp.matrix(1).text == '&&'
    assert exp.matrix(1).type == TokenType.And
    assert exp.matrix(2).name == 'Anonymous'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).text == '$*.*.CRONGO'
    assert exp.matrix(2, 0).type == TokenType.Variable
    assert exp.matrix(2, 1).text == '=='
    assert exp.matrix(2, 1).type == TokenType.Eq
    assert exp.matrix(2, 2).text == "'Y'"
    assert exp.matrix(2, 2).type == TokenType.Identifier
    assert exp.matrix(3).text == '?'
    assert exp.matrix(3).type == TokenType.Question
    assert exp.matrix(4).name == 'Anonymous'
    assert exp.matrix(4).type == TokenType.Expression
    assert exp.matrix(4, 0).text == '$*.*.*'
    assert exp.matrix(4, 0).type == TokenType.Variable
    assert exp.matrix(4, 1).text == '=='
    assert exp.matrix(4, 1).type == TokenType.Eq
    assert exp.matrix(4, 2).text == '\'\''
    assert exp.matrix(4,2).type == TokenType.Identifier
    assert exp.matrix(5).text == ':'
    assert exp.matrix(5).type == TokenType.Colon
    assert exp.matrix(6).text == 'true'
    assert exp.matrix(6).type == TokenType.TRUE


def case28():
    rule = StringIO("mustAnswer($*.*.CRONGO)&&$*.*.CRONGO=='N'?$*.*.*!='':true")
    exp = evals(rule)
    assert exp.name == 'TernaryOperator'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'mustAnswer'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).text == '$*.*.CRONGO'
    assert exp.matrix(0, 0).type == TokenType.Variable
    assert exp.matrix(1).text == '&&'
    assert exp.matrix(1).type == TokenType.And
    assert exp.matrix(2).name == 'Anonymous'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).text == '$*.*.CRONGO'
    assert exp.matrix(2, 0).type == TokenType.Variable
    assert exp.matrix(2, 1).text == '=='
    assert exp.matrix(2, 1).type == TokenType.Eq
    assert exp.matrix(2, 2).text == "'N'"
    assert exp.matrix(2, 2).type == TokenType.Identifier
    assert exp.matrix(3).text == '?'
    assert exp.matrix(3).type == TokenType.Question
    assert exp.matrix(4).name == 'Anonymous'
    assert exp.matrix(4).type == TokenType.Expression
    assert exp.matrix(4, 0).text == '$*.*.*'
    assert exp.matrix(4, 0).type == TokenType.Variable
    assert exp.matrix(4, 1).text == '!='
    assert exp.matrix(4, 1).type == TokenType.NEq
    assert exp.matrix(4, 2).text == '\'\''
    assert exp.matrix(4,2).type == TokenType.Identifier
    assert exp.matrix(5).text == ':'
    assert exp.matrix(5).type == TokenType.Colon
    assert exp.matrix(6).text == 'true'
    assert exp.matrix(6).type == TokenType.TRUE


def case29():
    rule = StringIO("""autoIncrease(1,1,1,1)""")
    exp = evals(rule)
    assert exp.name == 'autoIncrease'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).text == '1'
    assert exp.matrix(0).type == TokenType.RealNumber
    assert exp.matrix(1).text == '1'
    assert exp.matrix(1).type == TokenType.RealNumber
    assert exp.matrix(2).text == '1'
    assert exp.matrix(2).type == TokenType.RealNumber
    assert exp.matrix(3).text == '1'
    assert exp.matrix(3).type == TokenType.RealNumber


def case30():
    rule = StringIO("""$*.*.*=='Y'&&mustAnswer($*.Demographics.AGE)?toNum($*.Demographics.AGE)>=18:true""")
    exp = evals(rule)
    assert exp.name == 'TernaryOperator'
    assert exp.type == TokenType.Expression
    assert exp.matrix(0).name == 'Anonymous'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).name == 'Anonymous'
    assert exp.matrix(0, 0).type == TokenType.Expression
    assert exp.matrix(0, 0, 0).text == '$*.*.*'
    assert exp.matrix(0, 0, 0).type == TokenType.Variable
    assert exp.matrix(0, 0, 1).text == '=='
    assert exp.matrix(0, 0, 1).type == TokenType.Eq
    assert exp.matrix(0, 0, 2).text == '\'Y\''
    assert exp.matrix(0, 0, 2).type == TokenType.Identifier
    assert exp.matrix(0, 1).text == '&&'
    assert exp.matrix(0, 1).type == TokenType.And
    assert exp.matrix(0, 2).name == 'mustAnswer'
    assert exp.matrix(0, 2).type == TokenType.Expression
    assert exp.matrix(0, 2, 0).text == '$*.Demographics.AGE'
    assert exp.matrix(0, 2, 0).type == TokenType.Variable
    assert exp.matrix(1).text == '?'
    assert exp.matrix(1).type == TokenType.Question
    assert exp.matrix(2).name == 'Anonymous'
    assert exp.matrix(2).type == TokenType.Expression
    assert exp.matrix(2, 0).name == 'toNum'
    assert exp.matrix(2, 0).type == TokenType.Expression
    assert exp.matrix(2, 0, 0).text == '$*.Demographics.AGE'
    assert exp.matrix(2, 0, 0).type == TokenType.Variable
    assert exp.matrix(2, 1).text == '>='
    assert exp.matrix(2, 1).type == TokenType.Ge
    assert exp.matrix(2, 2).text == '18'
    assert exp.matrix(2, 2).type == TokenType.RealNumber
    assert exp.matrix(3).text == ':'
    assert exp.matrix(3).type == TokenType.Colon
    assert exp.matrix(4).text == 'true'
    assert exp.matrix(4).type == TokenType.TRUE


def case31():
    rule = StringIO("""a?1:2""")
    exp = evals(rule)
    assert exp.name == 'TernaryOperator'
    assert exp.matrix(0).text == 'a'
    assert exp.matrix(0).type == TokenType.Identifier
    assert exp.matrix(1).text == '?'
    assert exp.matrix(1).type == TokenType.Question
    assert exp.matrix(2).text == '1'
    assert exp.matrix(2).type == TokenType.RealNumber
    assert exp.matrix(3).text == ':'
    assert exp.matrix(3).type == TokenType.Colon
    assert exp.matrix(4).text == '2'
    assert exp.matrix(4).type == TokenType.RealNumber


def case32():
    rule = StringIO("a()")
    exp = evals(rule)
    assert exp.name == 'a'
    assert exp.type == TokenType.Expression
    assert len(exp.body) == 0


def case33():
    rule = StringIO("""a()?1:b()""")
    exp = evals(rule)

    assert exp.name == 'TernaryOperator'
    assert exp.matrix(0).name == 'a'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(1).text == '?'
    assert exp.matrix(1).type == TokenType.Question
    assert exp.matrix(2).text == '1'
    assert exp.matrix(2).type == TokenType.RealNumber
    assert exp.matrix(3).text == ':'
    assert exp.matrix(3).type == TokenType.Colon
    assert exp.matrix(4).name == 'b'
    assert exp.matrix(4).type == TokenType.Expression

def case34():
    rule = StringIO("""a(c())?1:b(d())""")
    exp = evals(rule)

    assert exp.name == "TernaryOperator"
    assert exp.matrix(0).name == 'a'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).name == 'c'
    assert exp.matrix(0, 0).type == TokenType.Expression
    assert exp.matrix(1).text == '?'
    assert exp.matrix(1).type == TokenType.Question
    assert exp.matrix(2).text == '1'
    assert exp.matrix(2).type == TokenType.RealNumber
    assert exp.matrix(3).text == ':'
    assert exp.matrix(3).type == TokenType.Colon
    assert exp.matrix(4).name == 'b'
    assert exp.matrix(4).type == TokenType.Expression
    assert exp.matrix(4, 0).name == 'd'
    assert exp.matrix(4, 0).type == TokenType.Expression
    # print(exp)


def case35():
    rule = StringIO("a(c())?1:b(d(1,2),x())")
    exp = evals(rule)
    assert exp.name == "TernaryOperator"
    assert exp.matrix(0).name == 'a'
    assert exp.matrix(0).type == TokenType.Expression
    assert exp.matrix(0, 0).name == 'c'
    assert exp.matrix(0, 0).type == TokenType.Expression
    assert exp.matrix(1).text == '?'
    assert exp.matrix(1).type == TokenType.Question
    assert exp.matrix(2).text == '1'
    assert exp.matrix(2).type == TokenType.RealNumber
    assert exp.matrix(3).text == ':'
    assert exp.matrix(3).type == TokenType.Colon
    assert exp.matrix(4).name == 'b'
    assert exp.matrix(4).type == TokenType.Expression
    assert exp.matrix(4, 0).name == 'd'
    assert exp.matrix(4, 0).type == TokenType.Expression
    assert exp.matrix(4, 0, 0).text == '1'
    assert exp.matrix(4, 0, 0).type == TokenType.RealNumber
    assert exp.matrix(4, 0, 1).text == '2'
    assert exp.matrix(4, 0, 1).type == TokenType.RealNumber
    assert exp.matrix(4, 1).name == 'x'
    assert exp.matrix(4, 1).type == TokenType.Expression

def test_interpreter2():
    case35()


def test_interpreter():
    case0()
    case1()
    case2()
    case3()
    case4()
    case5()
    case6()
    case7()
    case8()
    case9()
    case10()
    case11()
    case12()
    case13()
    case14()
    case15()
    case16()
    case17()
    case18()
    case19()
    case20()
    case21()
    case22()
    case23()
    case24()
    case25()
    case26()
    case27()
    case28()
    case29()
    case30()
    case31()
    case32()
    case33()
    case34()
    case35()


from io import StringIO

from edkrule.interpreter.lexer.lexer import Lexer
from edkrule.interpreter.lexer.tokens.token import Token
from edkrule.interpreter.parse.statement_parser.ternary_statement_parser import TernaryStatementParser

from edkrule.interpreter.parse.statement import Statement
from edkrule.interpreter.parse.statement_parser.bracket_statement_parser import BracketStatementParser
from edkrule.interpreter.parse.statement_parser.colon_statement_parser import ColonStatementParser
from edkrule.interpreter.parse.statement_parser.end_statement_parser import EndStatementParser
from edkrule.interpreter.parse.statement_parser.function_statement_parser import FunctionStatementParser
from edkrule.interpreter.parse.statement_parser.normal_statement_parser import NormalStatementParser
from edkrule.interpreter.parse.statement_parser.question_statement_parser import QuestionStatementParser
from edkrule.utils.stack import Stack


class Parser:
    def __init__(self, parse_tokens: list[Token]):
        self.statement = Statement()
        self.body = parse_tokens
        self.index = 0
        self.ternary_stack = Stack()

    def parse2(self):
        while self.index <= len(self.body) - 1:
            for exp in [FunctionStatementParser(self),
                        BracketStatementParser(self),
                        QuestionStatementParser(self),
                        ColonStatementParser(self),
                        NormalStatementParser(self),
                        EndStatementParser(self)
                        ]:
                if exp.accept(): continue
        return self

    def parse(self):
        while self.index <= len(self.body) - 1:
            for exp in [FunctionStatementParser(self),
                        BracketStatementParser(self),
                        TernaryStatementParser(self),
                        NormalStatementParser(self),
                        EndStatementParser(self)
                        ]:
                if exp.accept(): continue
        return self


def case1():
    s = StringIO("a(1+2+b(2))")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse2()
    print(es.statement)


def case0():
    s = StringIO("b(2,3+1,4)")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()
    print(es.statement)


def case2():
    s = StringIO("b(2,3+1,4)")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()
    print(es.statement)


def case3():
    s = StringIO("date((1+2)+3==(4+5), year(\"xxxx\"), 9+10)")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()
    print(es.statement)


def case4():
    s = StringIO("a+b==1?2+3:4+1+2")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()
    print(es.statement)


def case5():
    s = StringIO("taDate(a+b==1?2+3:4+1+2,111)")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()
    print(es.statement)


def case6():
    s = StringIO("taDate(a+b==1?(2+3):4+1+2,111)")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()
    print(es.statement)


def case7():
    s = StringIO("taDate(a+b==1?(data(2+3)?(x+1):ddd):4+1+2,111)")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()
    print(es.statement)


def case8():
    s = StringIO("a+b==1?(data(2+3)?(x+10):ddd):4+1+2+111")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()
    print(es.statement)


def case9():
    s = StringIO(
        "autoValue((toNum($*.*.EGORRES.881)!=''&&toNum($*.*.EGORRES.879)!=''?RoundN(toNum($*.*.EGORRES.879)/Math.pow((toNum($*.*.EGORRES.881)/1000), 0.33), \"1\"):\"\"), true)")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()
    print(es.statement)


def case10():
    s = StringIO("$*.*.*=='Y'&&mustAnswer($*.Demographics.AGE)?toNum($*.Demographics.AGE)>=18:true")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()
    print(es.statement)


def case11():
    s = StringIO("1==1?2==2?3:1:1")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()
    print(es.statement)

def case12():
    s = StringIO("1==2?2+1:3==4?4:5")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()

def case13():
    s = StringIO("a==1?true:(x==1?y:z)")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()
    print(es)

def case14():
    s = StringIO("a(c())?1:b(d())")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()

def case15():
    s = StringIO("a(c())?1:b(d(1,2),x())")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()
    print(es)

def case16():
    # s = StringIO("autoincrease(1,1)==3?autoincrease(1,1):autoincrease(autoincrease(1,1),1),3+1,4+1)")
    s = StringIO("a?autoincrease(1,1):autoincrease(autoincrease(1,1),1),3)")
    # s = StringIO("a(c())?1:b(d(1,2),x())")
    l = Lexer()
    l.tokenize(s)
    es = Parser(l.token_list)
    es.parse()
    print(es)


if __name__ == "__main__":
   case16()


from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.interpreter.lexer.tokens.token import Token


class FalseToken:
    @staticmethod
    def accept(lexer, char: str):
        if Character.isalpha(char) and Character.isf(char):
            lexer.token_text += char
            lexer.state = DfaState.False1
            return True
        # elif Character.isalpha(char) and Character.isa(char) and lexer.state == DfaState.False1:
        #     lexer.token_text += char
        #     lexer.state = DfaState.False2
        #     return True
        # elif Character.isalpha(char) and Character.isl(char) and lexer.state == DfaState.False2:
        #     lexer.token_text += char
        #     lexer.state = DfaState.False3
        #     return True
        # elif Character.isalpha(char) and Character.iss(char) and lexer.state == DfaState.False3:
        #     lexer.token_text += char
        #     lexer.state = DfaState.False4
        #     return True
        # elif Character.isalpha(char) and Character.ise(char) and lexer.state == DfaState.False4:
        #     lexer.token_text += char
        #     lexer.state = DfaState.False5
        #     lexer.token_type = TokenType.FALSE
        #     return True
        return False



from lxml import etree

from edkrule.engine.constant import Constant




class XMLLoader:
    def __init__(self, path):
        self.xml_define = etree.parse(path)

    @property
    def namespace(self):
        return "{http://www.edkrule.org/XMLSchema}"

    def equal(self, tag: str, element_name: str) -> bool:
        return tag.replace(self.namespace, "").lower() == element_name

    def text(self, tag) -> str:
        return tag.replace(self.namespace, "").lower()

    def loader(self, engine_definition):
        root = self.xml_define.getroot()

        if root.attrib.get("parent") is not None:
            XMLLoader(root.attrib.get("parent")).loader(engine_definition)

        for e in self.xml_define.getroot():
            if self.equal(e.tag, Constant.DataTypes):
                self.data_type_definition(e, engine_definition)
            elif self.equal(e.tag, Constant.Expressions):
                self.expression_definition(e, engine_definition)
            elif self.equal(e.tag, Constant.Identifiers):
                self.identifier_definition(e, engine_definition)

    def class_info_load(self, engine_definition_type, element, engine_definition):
        package = element.attrib.get(Constant.Package)
        for e in element:
            engine_definition.set(engine_definition_type, e.attrib.get(Constant.Name).lower(), {
                Constant.Package: e.attrib.get(Constant.Package) or package,
                Constant.Class: e.attrib.get(Constant.Class)})

    def data_type_definition(self, element, engine_definition):
        self.class_info_load(Constant.DataTypes, element, engine_definition)

    def expression_definition(self, element, engine_definition):
        self.class_info_load(Constant.Expressions, element, engine_definition)

    def identifier_definition(self, element, engine_definition):
        self.class_info_load(Constant.Identifiers, element, engine_definition)




from io import StringIO

from edkrule.edk_rule import EdkRule
from edkrule.engine.constant import Constant
from edkrule.engine.engine import Engine
from edkrule.engine.finder.by import By
from edkrule.interpreter.expression.interpreter import Interpreter
from edkrule.interpreter.lexer.lexer import Lexer
from edkrule.interpreter.parse.parser import Parser


def case0():
    engine = Engine(
        definition_path="D:\\aws\\eclinical40_auto_testing\\edkrule\\src\\edkrule\\tests\\engine\\engine.xml")
    assert engine.get(Constant.DataTypes, 'realNumber').get(Constant.Package) == 'dp1'
    assert engine.get(Constant.DataTypes, 'realNumber').get(Constant.Class) == 'dc1'
    assert engine.get(Constant.DataTypes, 'VarIable').get(Constant.Package) == 'dp2'
    assert engine.get(Constant.DataTypes, 'VarIable').get(Constant.Class) == 'dc2'
    assert engine.get(Constant.Identifiers, 'xx').get(Constant.Package) == 'sp1'
    assert engine.get(Constant.Identifiers, 'xx').get(Constant.Class) == 'sc1'
    assert engine.get(Constant.Identifiers, 'yyy').get(Constant.Package) == 'sp2'
    assert engine.get(Constant.Identifiers, 'YYy').get(Constant.Class) == 'sc2'
    assert engine.get(Constant.Expressions, '+').get(Constant.Package) == 'ep1'
    assert engine.get(Constant.Expressions, '+').get(Constant.Class) == 'ec1'
    assert engine.get(Constant.Expressions, '-').get(Constant.Package) == 'ep2'
    assert engine.get(Constant.Expressions, '-').get(Constant.Class) == 'ec2'
    assert engine.get(Constant.Expressions, '||').get(Constant.Package) == 'ep'
    assert engine.get(Constant.Expressions, '||').get(Constant.Class) == 'ec3'
    assert engine.get(Constant.Expressions, '&&').get(Constant.Package) == 'ep'
    assert engine.get(Constant.Expressions, '&&').get(Constant.Class) == 'ec4'

    assert engine.definition.identifier("dasdsa").get(Constant.Package) == 'sp'
    assert engine.definition.identifier("dasdsa").get(Constant.Class) == 'ic'


def case1():
    engine = Engine(
        definition_path="D:\\aws\\eclinical40_auto_testing\\edkrule\\src\\edkrule\\tests\\engine\\engine0.xml")
    assert engine.get(Constant.DataTypes, 'realNumber').get(Constant.Package) == 'cdp1'
    assert engine.get(Constant.DataTypes, 'realNumber').get(Constant.Class) == 'cdc1'
    assert engine.get(Constant.DataTypes, 'VarIable').get(Constant.Package) == 'dp2'
    assert engine.get(Constant.DataTypes, 'VarIable').get(Constant.Class) == 'dc2'
    assert engine.get(Constant.Identifiers, 'xx').get(Constant.Package) == 'csp1'
    assert engine.get(Constant.Identifiers, 'xx').get(Constant.Class) == 'sc1'
    assert engine.get(Constant.Identifiers, 'yyy').get(Constant.Package) == 'sp2'
    assert engine.get(Constant.Identifiers, 'YYy').get(Constant.Class) == 'sc2'
    assert engine.get(Constant.Expressions, '+').get(Constant.Package) == 'ep1'
    assert engine.get(Constant.Expressions, '+').get(Constant.Class) == 'ec1'
    assert engine.get(Constant.Expressions, '-').get(Constant.Package) == 'ep2'
    assert engine.get(Constant.Expressions, '-').get(Constant.Class) == 'ec2'
    assert engine.get(Constant.Expressions, '||').get(Constant.Package) == 'cep3'
    assert engine.get(Constant.Expressions, '||').get(Constant.Class) == 'ec3'
    assert engine.get(Constant.Expressions, '&&').get(Constant.Package) == 'ep'
    assert engine.get(Constant.Expressions, '&&').get(Constant.Class) == 'ec4'
    assert engine.definition.identifier("dasdsa").get(Constant.Package) == 'sp'
    assert engine.definition.identifier("dasdsa").get(Constant.Class) == 'ic'


def evals(stringio):
    l = Lexer()
    l.tokenize(stringio)
    es = Parser(l.token_list)
    es.parse()
    exp = Interpreter().interpret(es.statement)
    return exp


def load_engine():
    engine = Engine(
        definition_path="D:\\aws\\eclinical40_auto_testing\\edkrule\\src\\edkrule\\tests\\engine\\engine2.xml")
    return engine


def case2():
    engine = load_engine()
    rule = StringIO("111")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == 111
    rule = StringIO("a111")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == "a111"
    rule = StringIO("true")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == True
    rule = StringIO("false")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == False


def case3():
    engine = load_engine()
    rule = StringIO("1+1+3")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == 5


def case4():
    engine = load_engine()
    rule = StringIO("1==1?2:3")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == 2
    engine = load_engine()
    rule = StringIO("1==2?2+1:3+5")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == 8


def case5():
    engine = load_engine()
    rule = StringIO("1==2?2+1:3==4?4:5")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == 5


def case6():
    engine = load_engine()
    rule = StringIO("toDate()?2+1:3==4?4:5")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == 5


def case7():
    engine = load_engine()
    rule = StringIO("autoincrease(1,2,3,4)")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == 10
    engine = load_engine()
    rule = StringIO("autoincrease(autoincrease(1,1),autoincrease(autoincrease(1,1),1),3+1,4+1)")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == 14


def case8():
    engine = load_engine()
    rule = StringIO("$*.*.*")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == "$*.*.*"


def case9():
    engine = load_engine()
    rule = StringIO(
        "autoincrease(1,1)==2?autoincrease(1,1):autoincrease(autoincrease(1,1),autoincrease(autoincrease(1,1),1),3+1,4+1)")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == 2
    engine = load_engine()
    rule = StringIO(
        "autoincrease(1,1)==3?autoincrease(1,1):autoincrease(autoincrease(1,1),autoincrease(autoincrease(1,1),1),3+1,4+1)")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == 14


def case10():
    engine = load_engine()
    rule = StringIO(
        "autoincrease(autoincrease(1,1)==2?autoincrease(1,1):autoincrease(autoincrease(1,1),autoincrease(autoincrease(1,1),1),3+1,4+1),autoincrease(1,1)==3?autoincrease(1,1):autoincrease(autoincrease(1,1),autoincrease(autoincrease(1,1),1),3+1,4+1))")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == 16


def case11():
    engine = load_engine()
    rule = StringIO("2!=1")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == True
    engine = load_engine()
    rule = StringIO("2>=1")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == True
    engine = load_engine()
    rule = StringIO("2>=2")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == True
    engine = load_engine()
    rule = StringIO("1<=2")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == True
    engine = load_engine()
    rule = StringIO("1<=1")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == True


def case12():
    engine = load_engine()
    rule = StringIO("true && false")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == False
    engine = load_engine()
    rule = StringIO("true || false")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == True


def case13():
    engine = load_engine()
    rule = StringIO("1-2")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == -1
    engine = load_engine()
    rule = StringIO("1*2")
    exp = evals(rule)
    exp.engine = engine
    r = exp.run()
    assert r == 2
    r = EdkRule.run(rule_string="1/2")
    assert r.get() == 0.5


def case14():
    engine = load_engine()
    rule = StringIO("1-2")
    exp = evals(rule)
    exp.engine = engine
    exp.run()
    rid = exp.find(By.RuleFragment, "1").rid
    assert engine.cache.get(rid) == 1


def case15():
    # r = EdkRule.run(rule_string="1/2")
    # print(r.track())
    r = EdkRule.run(rule_string="autoincrease(autoincrease(1,1)==2?autoincrease(1,1):autoincrease(autoincrease(1,1),autoincrease(autoincrease(1,1),1),3+1,4+1),autoincrease(1,1)==3?autoincrease(1,1):autoincrease(autoincrease(1,1),autoincrease(autoincrease(1,1),1),3+1,4+1))")
    # r = EdkRule.run(rule_string="autoincrease(1,2)")
    print(r.get())
    print(r.track())

def test_debug():
    rule = StringIO("""condition(mustAnswer(toNum($*.*.*)), isRange($*.*.*, 40,160))""")
    exp = evals(rule)
    print(exp.text)
    rid = exp.find(By.RuleFragment, rule.getvalue()).rid
    assert rid != None

def test_cases():
    case0()
    case1()
    case2()
    case3()
    case4()
    case5()
    case6()
    case7()
    case8()
    case9()
    case10()
    case11()
    case12()
    case13()
    case14()
    case15()



from edkrule.edk_rule import EdkRule







if __name__ == '__main__':
    # rule_string = """autoValue(RoundN(sum(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG, $*.Target Lesions Assessment (Details) (Screening).TLLOC != 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU == "CM") , multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLLONG, $*.Target Lesions Assessment (Details) (Screening).TLLOC != 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU == "MM"),1/10) , getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT, $*.Target Lesions Assessment (Details) (Screening).TLLOC == 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU== "CM") , multiply(getSumOfItemInLog($*.Target Lesions Assessment (Details) (Screening).TLSHORT, $*.Target Lesions Assessment (Details) (Screening).TLLOC == 11 && $*.Target Lesions Assessment (Details) (Screening).TLDIAU== "MM"),1/10)), 0.01), true)"""
    rule_string = """a&&b?2+3:4+1+2"""
    # rule_string = """ a()!=""&& b()!=""?dateDiff($C1D1.ONC-392 Administration.ECSTDAT+" "+$C1D1.ONC-392 Administration.ECSTTIM+":00",$*.*.LBDAT+" "+$*.*.*+":00","m")>0:true"""
    expression = EdkRule.expression(rule_string)

    data = expression.tree_data()
    EdkRule.draw(rule_string, "rule1999.html")


import enum


class TokenType(enum.Enum):
    Expression = 1000
    Statement = 1001
    Initial = 0
    Identifier = 1
    Assignment = 2
    Eq = 3
    NEq = 4
    Lt = 5
    Le = 6
    Gt = 7
    Ge = 8
    Variable = 9
    Lp = 10
    Rp = 11
    Dq = 12
    Sq = 13
    Colon = 14
    Plus = 15
    Minus = 16
    RealNumber = 17
    ByteAnd = 18
    And = 19
    ByteOr = 20
    Or = 21
    TRUE = 22
    FALSE = 23
    Blank = 24
    DPlus = 25
    PlusEq = 26
    DMinus = 27
    MinusEq = 28
    Comma = 29
    Question = 30
    Ep = 31
    Nidentity = 32
    AwaysEq = 33
    EqPlus = 34
    EqMinus = 35
    Divide = 36
    Multipy = 37
    Dot = 38



from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class NeqToken:
    @staticmethod
    def accept(lexer, char: str):
        if lexer.state == DfaState.NEq:
            lexer.state = DfaState.NEq
            lexer.token_type = TokenType.NEq
            lexer.token_text += char
            return True
        return False



from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class EqState:
    @staticmethod
    def accept(lexer, char: str, state: DfaState):
        if state == DfaState.Eq and Character.isassign(char):
            lexer.token_text += char
            lexer.state = DfaState.AwaysEq
            return True
        elif state == DfaState.Eq:
            lexer.token_type = TokenType.Eq
            lexer.init_token(char)
            return True
        return False



from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class Token:
    def __init__(self, t_type: TokenType = TokenType.Initial, t_text: str = ""):
        self._type: TokenType = t_type
        self._text: str = t_text

    @property
    def type(self): return self._type

    @type.setter
    def type(self, value): self._type = value

    @property
    def text(self): return self._text

    @text.setter
    def text(self, value): self._text = value

    def append(self, char: str):
        self._text = self._text + char



import importlib

from edkrule.engine.constant import Constant


class EngineDefinition:
    def __init__(self):
        self._definition = {
            Constant.DataTypes: {},
            Constant.Identifiers: {},
            Constant.Expressions: {}
        }

    def set(self, engine_definition_type, key, value):
        self._definition[engine_definition_type][key] = value

    def get(self, engine_definition_type, name):
        return self._definition[engine_definition_type].get(name)

    def get_class(self, engine_definition_type, name):
        class_info = None
        if engine_definition_type == Constant.DataTypes:
            class_info = self.data_type(name)
        elif engine_definition_type == Constant.Expressions:
            class_info = self.expression(name)
        elif engine_definition_type == Constant.Identifiers:
            class_info = self.identifier(name)
        if class_info is None: return None
        return self.class_loader(class_info)

    def class_loader(self, class_info):
        pkg, clazz_name = class_info.get(Constant.Package), class_info.get(Constant.Class).split(".")
        module = importlib.import_module(pkg+f".{clazz_name[0]}")
        entry_clazz = getattr(module, clazz_name[1])
        return entry_clazz

    def data_type(self, name):
        return self._definition[Constant.DataTypes].get(name)

    def identifier(self, name):
        class_info = self._definition[Constant.Identifiers].get(name)
        if class_info is None:
            class_info = self._definition[Constant.Identifiers].get(Constant.DefaultIdentifier)
        return class_info

    def expression(self, name):
        return self._definition[Constant.Expressions].get(name)



from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class FalseState:

    @staticmethod
    def accept(lexer, char: str, state: DfaState):
        if state == DfaState.False1:
            lexer.token_text += char
            lexer.state = DfaState.False2 if Character.isa(char) else DfaState.Identifier
            # lexer.state = DfaState.False2
            return True
        elif state == DfaState.False2:
            lexer.token_text += char
            lexer.state = DfaState.False3 if Character.isl(char) else DfaState.Identifier
            # lexer.state = DfaState.False3
            return True
        elif state == DfaState.False3:
            lexer.token_text += char
            lexer.state = DfaState.False4 if Character.iss(char) else DfaState.Identifier
            # lexer.state = DfaState.False4
            return True
        elif state == DfaState.False4:
            lexer.token_text += char
            lexer.state = DfaState.False5 if Character.ise(char) else DfaState.Identifier
            # lexer.state = DfaState.False5
            return True
        elif state == DfaState.False5:
            if Character.isseparators(char) or Character.iscolon(char):
                lexer.token_type = TokenType.FALSE
                lexer.init_token(char)
            else:
                lexer.token_text += char
                lexer.state = DfaState.Identifier
            return True

        return False



from io import StringIO
from pathlib import Path

from edkrule.engine.engine import Engine
from edkrule.engine.result import Result
from edkrule.interpreter.expression.interpreter import Interpreter
from edkrule.interpreter.lexer.lexer import Lexer
from edkrule.interpreter.parse.parser import Parser
from edkrule.utils.output.by import By
from edkrule.utils.rule_show import RuleShow


# TODO 导出 语法路径 HTML
# TODO 规则阅读器 APP
# TODO 词法解析的时候，不应该把Math.power()的形式解析成Math.power, ( , )
# TODO 这种函数的判断应该放在语法分析
# TODO 支持正则表达式
class EdkRule:
    @staticmethod
    def lexer(rule_string: str):
        """
        返回规则词法分析的结果
        :param rule_string:
        :type rule_string:
        :return:
        :rtype:
        """
        l = Lexer()
        l.tokenize(StringIO(rule_string))
        return l.token_list

    @staticmethod
    def parser(rule_string: str):
        """
        返回规则语法分析的结果
        :param rule_string:
        :type rule_string:
        :return:
        :rtype:
        """
        es = Parser(EdkRule.lexer(rule_string))
        es.parse()
        return es.statement

    @staticmethod
    def expression(rule_string: str):
        """
        返回规则表达式分析的结果
        :param rule_string:
        :type rule_string:
        :return:
        :rtype:
        """
        statement = EdkRule.parser(rule_string)
        exp = Interpreter().interpret(statement)
        exp.origin_statement(statement)
        return exp

    @staticmethod
    def run(rule_string: str, engine_path: str = "", use_sys=True, none_default = False) -> Result:
        """
        运行规则
        传参方式 1.
            engine_path = “” and use_sys=True 只是用系统默认规则引擎
        传参方式 2.
            engine_path = “xxxxx” and use_sys=True 同时加载传入的引擎配置和系统默认规则引擎
            如果有相同的名字的，加载传入的引擎配置会覆盖系统默认的配置
        传参方式 3.
            engine_path = “xxxxx” and use_sys=False 只加载传入的引擎配置
        :param rule_string: 规则字符串
        :type rule_string:
        :param engine_path: 加载的 引擎配置的 XML 路径
        :type engine_path:
        :param use_sys: 是否加载系统默认的规则
        :type use_sys:
        :return:
        :rtype:
        """
        try:
            engine = Engine(definition_path=engine_path, use_sys=use_sys)
            rule = EdkRule.expression(rule_string)
            rule.engine = engine
            rule.run()
            return Result(rule)
        except Exception as e:
            if none_default: return None
            raise e

    @staticmethod
    def draw(rule_string: str, output_path: Path):
        """
        画出规则的语法树
        :param rule_string:
        :type rule_string:
        :param output_path:
        :type output_path:
        :return:
        :rtype:
        """
        expression = EdkRule.expression(rule_string)
        data = expression.tree_data()

        RuleShow().output(By.Html, output_path, data)


from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class TrueToken:
    @staticmethod
    def accept(lexer, char: str):
        if Character.isalpha(char) and Character.ist(char):
            lexer.state = DfaState.True1
            lexer.token_text += char
            return True
        # elif Character.isalpha(char) and Character.isr(char) and lexer.state == DfaState.True1:
        #     lexer.state = DfaState.True2
        #     lexer.token_text += char
        #     return True
        # elif Character.isalpha(char) and Character.isu(char) and lexer.state == DfaState.True2:
        #     lexer.state = DfaState.True3
        #     lexer.token_text += char
        #     return True
        # elif Character.isalpha(char) and Character.ise(char) and lexer.state == DfaState.True3:
        #     lexer.state = DfaState.True4
        #     lexer.token_text += char
        #     lexer.token_type = TokenType.TRUE
        #     return True
        return False



import abc
import pathlib
from pathlib import Path


class Printer(abc.ABC):
    @staticmethod
    def default_jinja_template():
        return pathlib.Path(__file__).parent.joinpath(pathlib.Path("tree-vertical.html.jinja"))

    @abc.abstractmethod
    def output(self, out_path: Path, data, output_format): ...


from io import StringIO

from edkrule.interpreter.lexer.lexer import Lexer
from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.utils.format.formater import Formater


class BlankFormater(Formater):
    def initialize(self, string) -> str:
        """
        格式化，去除空格
        :param string:
        :type string:
        :return:
        :rtype:
        """
        lexer = Lexer()
        lexer.tokenize(StringIO(string))
        text = "".join([e.text for e in list(filter(lambda e: e.type != TokenType.Blank, lexer.token_list))])
        return text



from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class EpState:

    @staticmethod
    def accept(lexer, char: str, state: DfaState):
        if Character.isexclamationpoint(char):
            lexer.token_text += char
            lexer.token_type = TokenType.Ep
            lexer.state = DfaState.Ep
            return True
            # lexer.init_token(char)
        elif lexer.state == DfaState.Ep and Character.isassign(char):
            lexer.token_text += char
            # lexer.token_type = TokenType.NEq
            lexer.state = DfaState.NEq
            return True
        return False




from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState


class EqToken:
    @staticmethod
    def accept(lexer, char: str):
        if lexer.state == DfaState.Assignment and Character.isassign(char):
            lexer.token_text += char
            lexer.state = DfaState.Eq
            return True
        return False



from edkrule.interpreter.lexer.token_type import TokenType
from edkrule.interpreter.parse.statement import Statement
from edkrule.interpreter.parse.statement_parser.base_statement_parser import AbcStatementParser


class NormalStatementParser(AbcStatementParser):

    def __init__(self, parser):
        super().__init__(parser)
        self.statement = Statement()

    def accept(self):
        if self.parser.index <= len(self.parser.body) - 1:
            if self.token().type == TokenType.Identifier and self.parser.index < len(
                    self.parser.body) - 1 and self.token(offset=1).type == TokenType.Lp:
                # FunctionStatement 改函数跳出循环
                return False
            if self.token().type in [TokenType.Plus, TokenType.Comma, TokenType.Eq,
                                     TokenType.Minus, TokenType.DPlus, TokenType.DMinus,
                                     TokenType.Multipy, TokenType.Divide,
                                     TokenType.And, TokenType.Or, TokenType.Eq,
                                     TokenType.Ep, TokenType.NEq, TokenType.Nidentity,
                                     TokenType.EqPlus, TokenType.MinusEq, TokenType.EqMinus,
                                     TokenType.PlusEq, TokenType.Assignment
                                     ]:
                self.parser.statement.add(self.token())
            elif self.token().type in [TokenType.Question, TokenType.Colon, TokenType.Lp, TokenType.Rp]:
                return False
            else:
                self.statement.add(self.token())
                self.parser.statement.add(self.statement)
            self.move()
        return True



from edkrule.interpreter.lexer.character import Character
from edkrule.interpreter.lexer.dfa_state import DfaState
from edkrule.interpreter.lexer.token_type import TokenType


class PlusState:
    @staticmethod
    def accept(lexer, char: str, state: DfaState):
        if state == DfaState.Plus:
            if Character.isplus(char):
                lexer.token_text += char
                lexer.state = DfaState.DPlus
            elif Character.isassign(char):
                lexer.token_text += char
                lexer.state = DfaState.PlusEq
            else:
                lexer.token_type = TokenType.Plus
                lexer.init_token(char)
            return True
        return False




from edkrule.engine.finder.finder import Finder
from edkrule.interpreter.lexer.tokens.token import Token


class FinderByRid(Finder):
    def find(self, expression, rid):
        if rid == expression.rid:
            return expression
        else:
            for body in expression.body:
                if type(body) != Token:
                    rule = body.find(FinderByRid(), rid)
                    if rule is not None: return rule
